
SetupFiles <- function(Source="..", Target=".", Platform=FALSE) {
  Binaries <- c("AlphaSim", "macs", "AlphaFormatter")
  if (Platform) {
    PlatformString <- paste0("_", system("uname", intern=TRUE))
  } else {
    PlatformString <- ""
  }
  for (Binary in Binaries) {
    system(command=paste0("ln -sf ", Source, "/", Binary, PlatformString, " ", Target, "/", Binary))
  }
  SpecTemplates <- c("AlphaSimSpecTemplate.txt")
  for (SpecTemplate in SpecTemplates) {
    system(command=paste0("cp -f ", Source, "/", SpecTemplate, " ", Target))
  }
}

MaCS <- function(MaCSParameters, n=NULL) {
  MaCSParameters <- paste0(MaCSParameters, " -s ", round(runif(n=1, min=1000, max=1000000)))
  System(Command=paste0("./macs ", MaCSParameters, " 1>output.txt 2>debug.txt"))
  System(Command="./AlphaFormatter")
}

SimulateBaseHaplotypes <- function(nChrom, Chrom=NULL, MaCSParameters, ...) {
  # unlink("Chromosomes", recursive=TRUE)
  dir.create("Chromosomes", showWarnings=FALSE)
  BaseDir <- setwd("Chromosomes")
  on.exit(setwd(BaseDir))
  if (is.null(Chrom)) {
    Chroms <- 1:nChrom
  } else {
    Chroms <- Chrom
  }
  for (Chrom in Chroms) {
    # Chrom <- 1
    cat("Chromosome ", Chrom, " START\n")
    Dir <- paste0("Chromosome", Chrom)
    dir.create(Dir, showWarnings=FALSE)
    OldDir <- setwd(Dir)
    system(command="ln -sf ../../macs")
    system(command="ln -sf ../../AlphaFormatter")
    MaCS(MaCSParameters=MaCSParameters, ...)
    cat("Chromosome ", Chrom, " END\n")
    setwd(OldDir)
  }
}

HandleChromosomesFolder <- function(ChromosomesFolder, MacsHaplotypesFile, CompressedBefore=FALSE) {
  # This function was done to avoid copying the whole Chromosomes folder (was to large in this particular case)
  # Instead of copying we link as much as we can
  Tmp <- dir(path=ChromosomesFolder)
  nChrom <- length(Tmp[grepl(pattern="Chromosome", x=Tmp)])
  dir.create("Chromosomes", showWarnings=FALSE)
  BaseDir <- setwd("Chromosomes")
  on.exit(setwd(BaseDir))
  for (Chrom in 1:nChrom) {
    Dir <- paste0("Chromosome", Chrom)
    dir.create(Dir, showWarnings=FALSE)
    OldDir <- setwd(Dir)
    System(Command=paste0("ln -sf ../../", ChromosomesFolder, "/Chromosome", Chrom, "/", MacsHaplotypesFile, " MacsHaplotypes.txt"))
    if (!CompressedBefore) {
      System(Command=paste0("ln -sf ../../", ChromosomesFolder, "/Chromosome", Chrom, "/SegSites.txt"))
      System(Command=paste0("ln -sf ../../", ChromosomesFolder, "/Chromosome", Chrom, "/SegSitesIncluded.txt"))
      System(Command=paste0("ln -sf ../../", ChromosomesFolder, "/Chromosome", Chrom, "/PhysicalMapInput.txt"))
    } else {
      System(Command=paste0("cp -f ../../", ChromosomesFolder, "/Chromosome", Chrom, "/SegSitesIncluded.txt ."))
      System(Command=paste0("cp -f SegSitesIncluded.txt SegSites.txt"))
      Tmp <- read.table(file=paste0("../../", ChromosomesFolder, "/Chromosome", Chrom, "/PhysicalMapIncluded.txt"), header=FALSE)
      cat(Tmp[[3]], "\n", file="PhysicalMapInput.txt")
    }
    setwd(OldDir)
  }
}

AlphaSim <- function(Seed) {
  if (is.null(Seed)) {
    Seed <- round(runif(n=1)*1000000)
  }
  cat(Seed, file="Seed.txt")
  System(Command="./AlphaSim")
}

SimulatePopulation <- function(nGen, nIndPerPop, nSire, nDam,
                               nChrom, nBaseHap,
                               UseOldSnpLocations=FALSE, UseOldQTNYesNo=FALSE, UseOldEffectsYesNo=FALSE, OldFolder=NULL, Corr=1,
                               Compression=FALSE, CompressionPct=NULL, CompressedBefore=FALSE, Seed=NULL, ChipAlleleFreq="0.00,1.00",
                               QtnAlleleFreq="0.00,0.25") {
  x <- read_lines("AlphaSimSpecTemplate.txt")
  x <- sub(pattern=",nGen", replacement=paste0(",", nGen), x=x)
  x <- sub(pattern=",nIndPerPop", replacement=paste0(",", nIndPerPop), x=x)
  x <- sub(pattern=",nSire", replacement=paste0(",", nSire), x=x)
  x <- sub(pattern=",nDam", replacement=paste0(",", nDam), x=x)
  x <- sub(pattern=",nChrom", replacement=paste0(",", nChrom), x=x)
  x <- sub(pattern=",nBaseHap", replacement=paste0(",", nBaseHap), x=x)
  x <- sub(pattern=",MinMaxAlleleFreqForSnp", replacement=paste0(",", ChipAlleleFreq), x=x)
  x <- sub(pattern=",MinMaxAlleleFreqForQtn", replacement=paste0(",", QtnAlleleFreq), x=x)
  if (!UseOldSnpLocations) {
    x <- sub(pattern=",UseOldSnpLocations", replacement=",No", x=x)
  } else {
    x <- sub(pattern=",UseOldSnpLocations", replacement=",Yes", x=x)
    nChip <- length(dir(paste0(OldFolder, "/AllIndividualsSnpChips/")))/2
    for (Chip in 1:nChip) {
      # Chip <- 1
      Data <- read_table(paste0(OldFolder, "/Chip", Chip, "SnpInformation.txt"))
      for (Chrom in 1:nChrom) {
        # Chrom <- 1
        Tmp <- filter(Data, ChromId == Chrom)
        #if (CompressedBefore) {
        #  UncompressedMap <- read.table(file=paste0("Chromosomes/Chromosome", Chrom, "/PhysicalMapUncompressed.txt"))
        #  names(UncompressedMap) <- c("OrigLocus", "Position", "Tmp")
        #  CompressedMap   <- read.table(file=paste0("Chromosomes/Chromosome", Chrom, "/PhysicalMapCompressed.txt"))
        #  names(CompressedMap) <- c("NewLocus", "OrigLocus", "Position", "Tmp")
        #  IncludedMap     <- read.table(file=paste0("Chromosomes/Chromosome", Chrom, "/PhysicalMapIncluded.txt"))
        #  names(IncludedMap) <- c("FinalLocus", "OrigLocus", "Position", "Tmp")
        #  MapFile <- paste0("Chromosomes/Chromosome", Chrom, "/PhysicalMapIncluded.txt")
        #  Map <- read.table(MapFile, header=FALSE)
        #}
        write_lines(Tmp[["SnpSeqIdOnChrom"]], path=paste0("Chip", Chip, "Chrom", Chrom, "Locations.txt"))
      }
    }
  }
  if (!UseOldQTNYesNo) {
    x <- sub(pattern=",UseOldQTNYesNo", replacement=",No", x=x)
  } else {
    x <- sub(pattern=",UseOldQTNYesNo", replacement=",Yes", x=x)
    DataU <- read_table(paste0(OldFolder, "/UnrestrictedQtnInformation.txt"))
    DataR <- read_table(paste0(OldFolder, "/RestrictedQtnInformation.txt"))
    for (Chrom in 1:nChrom) {
      # Chrom <- 1
      TmpU <- filter(DataU, ChromId == Chrom)
      TmpR <- filter(DataR, ChromId == Chrom)
      write_lines(TmpU[["QtnSeqIdOnChrom"]], path=paste0("UnrestrictedQtnLocations", Chrom, ".txt"))
      write_lines(TmpR[["QtnSeqIdOnChrom"]], path=paste0("RestrictedQtnLocations",   Chrom, ".txt"))
    }
  }
  if (!UseOldEffectsYesNo) {
    x <- sub(pattern=",UseOldEffectsYesNo", replacement=",No", x=x)
  } else {
    x <- sub(pattern=",UseOldEffectsYesNo", replacement=",Yes", x=x)
    file.copy(from=paste0(OldFolder, "/InitialTraitMeans.txt"), to="InitialTraitMeans.txt", overwrite=TRUE)
    dir.create("SimulatedData")
    file.copy(from="InitialTraitMeans.txt", to="SimulatedData/InitialTraitMeans.txt", overwrite=TRUE)
    Data <- read_table(paste0(OldFolder, "/UnrestrictedQtnAdditEffects.txt"), col_names=FALSE, skip=1)#, had issues reading in the file
    #col_types=cols(GlobalQtnId=col_integer(),
    #               ChromId=col_integer(),
    #               QtnSeqIdOnChrom=col_integer(),
    #               AddEffNormUnres1=col_double()))
    if (Corr < 1) {
      Data[["X4"]] <- rcorr(Data[["X4"]], r=Corr)
    }
    write_lines(Data[["X4"]], path="UnrestrictedQtnNormalAdditEffects.txt")
    write_lines(Data[["X4"]], path="UnrestrictedQtnGammaAdditEffects.txt") # Hack!

    Data <- read_table(paste0(OldFolder, "/RestrictedQtnAdditEffects.txt"), col_names=FALSE, skip=1)
    if (Corr < 1) {
      Data[["X4"]] <- rcorr(Data[["X4"]], r=Corr)
    }
    write_lines(Data[["X4"]], path="RestrictedQtnNormalAdditEffects.txt")
    write_lines(Data[["X4"]], path="RestrictedQtnGammaAdditEffects.txt") # Hack!

    Data <- read_table(paste0(OldFolder, "/UnrestrictedQtnAverAlleleSubstEffects.txt"), col_names=FALSE, skip=1)
    if (Corr < 1) {
      # Data[["X4"]] <- rcorr(Data[["X4"]], r=Corr)
    }
    write_lines(Data[["X4"]], path="UnrestrictedQtnNormalAverAlleleSubstEffects.txt")
    write_lines(Data[["X4"]], path="UnrestrictedQtnGammaAverAlleleSubstEffects.txt") # Hack!

    Data <- read_table(paste0(OldFolder, "/RestrictedQtnAverAlleleSubstEffects.txt"), col_names=FALSE, skip=1)
    if (Corr < 1) {
      # Data[["X4"]] <- rcorr(Data[["X4"]], r=Corr)
    }
    write_lines(Data[["X4"]], path="RestrictedQtnNormalAverAlleleSubstEffects.txt")
    write_lines(Data[["X4"]], path="RestrictedQtnGammaAverAlleleSubstEffects.txt") # Hack!

    Data <- read_table(paste0(OldFolder, "/UnrestrictedQtnDominEffects.txt"), col_names=FALSE, skip=1)
    if (Corr < 1) {
      # Data[["X4"]] <- rcorr(Data[["X4"]], r=Corr)
      # Data[["X5"]] <- rcorr(Data[["X5"]], r=Corr)
    }
    write_delim(Data[c("X4", "X5")], path="UnrestrictedQtnNormalDominEffects.txt", col_names=FALSE)
    write_delim(Data[c("X4", "X5")], path="UnrestrictedQtnGammaDominEffects.txt", col_names=FALSE) # Hack!

    Data <- read_table(paste0(OldFolder, "/RestrictedQtnDominEffects.txt"), col_names=FALSE, skip=1)
    if (Corr < 1) {
      # Data[["X4"]] <- rcorr(Data[["X4"]], r=Corr)
      # Data[["X5"]] <- rcorr(Data[["X5"]], r=Corr)
    }
    write_delim(Data[c("X4", "X5")], path="RestrictedQtnNormalDominEffects.txt", col_names=FALSE)
    write_delim(Data[c("X4", "X5")], path="RestrictedQtnGammaDominEffects.txt", col_names=FALSE) # Hack!
  }
  if (!Compression) {
    x <- sub(pattern=",GenomeCompression", replacement=",Off", x=x)
    x <- sub(pattern=",CompressionPct",    replacement=",100", x=x)
  } else {
    x <- sub(pattern=",GenomeCompression", replacement=",On", x=x)
    x <- sub(pattern=",CompressionPct",    replacement=paste0(",", CompressionPct), x=x)
  }
  write_lines(x, path="AlphaSimSpec.txt")
  # x <- read_lines("AlphaSimSpec.txt")
  AlphaSim(Seed=Seed)
}

rcorr <- function(x, r){
  if (FALSE) {
    x <- rnorm(100)
    y <- rcorr(x=x, r=.5)
    cor(x,y)
    y <- rcorr(x=x, r=1)
    cor(x,y)
  }
  r*x + rnorm(length(x), mean=0, sd=sqrt(1 - r**2))
}

