!ifort -O3 -heap-arrays -lmkl_blas95_lp64 -lmkl_lapack95_lp64 -lmkl_intel_lp64 -lmkl_intel_thread -lmkl_core -qopenmp -parallel -lpthread computer.f90  -o computer
program computer
 !$ use omp_lib
 implicit none
 integer::i,j,k,k0,k1,k2,io,un,nan,nsnp,nalleles,nsel,nanpop,npop,iter,msg=0
 integer,allocatable::chromo(:)
 character(len=30)::snpfile,mafa,title,mapfile
 real(kind=8)::maf,varg,vare,sum2pq,vars,t1
 real(kind=8),allocatable::genotype(:,:),freq(:),lhs(:,:),geno(:,:)
 logical::loptmem=.true.,lcentered=.false.
 logical,allocatable::selsnp(:)

 !Parameters
 call get_command_argument(1,value=title,status=io)
 call get_command_argument(2,value=snpfile,status=io)
 call get_command_argument(3,value=mafa,status=io)
 read(mafa,*)npop
 call get_command_argument(4,value=mafa,status=io)
 read(mafa,'(f16.8)')maf
 call get_command_argument(5,value=mapfile,status=io)

 if(lcentered)then
  write(*,'(a)')' Genotypes will be centered by 2p'
 else
  write(*,'(a)')' Genotypes will not be centered by 2p'
 endif

 !Genotypes
 open(newunit=un,file=snpfile,status='old',buffered='yes')
 call numlines(un,nan)
 call numcol(un,nsnp) 
 nsnp=nsnp-1
 allocate(genotype(nan,nsnp),freq(nsnp),selsnp(nsnp))
 genotype=-99.d0
 print'(a)',' Start reading the genotype file'
 do i=1,nan
  read(un,*,iostat=io)k,genotype(i,1:nsnp)
  if(io.ne.0)exit
 enddo
 close(un)
 write(*,'(a,i0)')' Number of animals: ',nan
 write(*,'(a,i0)')' Number of SNPs   : ',nsnp
 nanpop=nan/npop
 write(*,'(a,i0)')' Number of animals per mean: ',nanpop

 !read map file
 allocate(chromo(nsnp))
 chromo=0
 open(newunit=un,file=mapfile,buffered='yes')
 read(un,*)
 do i=1,nsnp
  read(un,*)k,chromo(i)
 enddo
 close(un)

 allocate(geno(nanpop,nsnp))
 do iter=1,npop
  write(mafa,'(i0)')iter

  geno=genotype((iter-1)*nanpop+1:iter*nanpop,:)
  !Computation of allele frequencies
  print'(/a)',' Computation of allele frequencies for pop '//adjustl(trim(mafa))
  freq=0.d0;selsnp=.true.
  nalleles=2.d0*nanpop
  sum2pq=0.d0
 
  j=0
  open(unit=100,file='sel_freq_'//adjustl(trim(title))//adjustl(trim(mafa))//'.dat',buffered='yes')
  !$omp parallel default(none) shared(j,sum2pq,nsnp,freq,geno,nalleles,selsnp,maf,lcentered) private(i,k,k0,k1,k2)
  !$omp do reduction(+:sum2pq,j)
  do i=1,nsnp
   freq(i)=sum(geno(:,i))/nalleles
   k0=count(geno(:,i).eq.0)
   k1=count(geno(:,i).eq.1)
   k2=count(geno(:,i).eq.2)
   k=k0*k1*k2
!   k=(count(geno(:,i).eq.0))*(count(geno(:,i).eq.1))*(count(geno(:,i).eq.2))
   if((freq(i)<=maf.or.freq(i)>=1.d0-maf).or.k.eq.0)then
    selsnp(i)=.false.
    geno(:,i)=0.d0
    freq(i)=0.d0
    j=j+1
   endif
   if(lcentered)geno(:,i)=geno(:,i)-2.d0*freq(i)
   sum2pq=sum2pq+2.d0*freq(i)*(1.d0-freq(i))
   !$omp critical
   write(100,*)i,freq(i),selsnp(i)
   !$omp end critical
  enddo
  !$omp end do
  !$omp end parallel
  close(100)
  nsel=count(selsnp.eq..true.)
  write(*,'(a,f0.3)')' Minor allele frequency   : ',maf
  write(*,'(a,i0)')' Number of selected SNPs  : ',nsel
  write(*,'(a,i0)')  ' Number of unselected SNPs: ',j
  write(*,'(a,f0.3)')' Sum of 2pq               : ',sum2pq
 
  open(newunit=k,file='sel_varused_'//adjustl(trim(title))//adjustl(trim(mafa)),buffered='yes',status='replace')
  do i=1,7
   write(k,*)
  enddo
  write(k,*)'nanimals',nanpop
  write(k,*)'nsnp',nsnp
  write(k,*)'nsel',nsel
  write(k,*)'sum2pq',sum2pq
  close(k)

  !Computation of r2
  if(.not.lcentered)then
   !$omp parallel do default(none) shared(nsnp,geno,freq) private(i)
   do i=1,nsnp
    geno(:,i)=geno(:,i)-2.d0*freq(i)
   enddo
   !$omp end parallel do
  endif
  allocate(lhs(nsnp,nsnp))
  call dgemm('t','n',nsnp,nsnp,nanpop,1.d0,geno,nanpop,geno,nanpop,0.d0,lhs,nsnp)
 
  open(unit=100,file='sel_adjsum_'//adjustl(trim(title))//adjustl(trim(mafa))//'.dat',buffered='yes')
  do i=1,nsnp
   write(100,'(i0,3(x,f20.10))')i,lhs(i,i),2*nanpop*freq(i)*(1.d0-freq(i)),freq(i)
  enddo
  close(100)

  !$omp parallel default(none) shared(nsnp,freq,lhs) private(i)
  !$omp do
  do i=1,nsnp
   freq(i)=sqrt(lhs(i,i))
  enddo
  !$omp end do
  !$omp do
  do i=1,nsnp
   do j=1,nsnp
    if((freq(i)*freq(j)).ne.0.d0)lhs(j,i)=(lhs(j,i)/(freq(i)*freq(j)))!**2
   enddo
  enddo
  !$omp end do
  !$omp end parallel 
  open(newunit=un,file='sel_r2_'//adjustl(trim(title))//adjustl(trim(mafa))//'.dat',buffered='yes',access='stream')
  write(un)lhs(1:nsnp,1:nsnp)
  close(un)

  if(msg.ne.0)then
   open(newunit=un,file='sel_r2_ascii_'//adjustl(trim(title))//'.dat',buffered='yes')
   do i=1,nsnp
    write(un,'(10000f16.8)')lhs(1:nsnp,i)
   enddo
   close(un)
  endif

  !$omp parallel do default(none) shared(chromo,nsnp,lhs) private(i,j)
  do i=1,nsnp
   do j=1,nsnp
    if(chromo(i).ne.chromo(j))then
     lhs(i,j)=0.d0
    endif
   enddo
  enddo
  !$omp end parallel do

  open(newunit=un,file='sel_r2c_'//adjustl(trim(title))//adjustl(trim(mafa))//'.dat',buffered='yes',access='stream')
  write(un)lhs(1:nsnp,1:nsnp)
  close(un)

  deallocate(lhs)
 enddo


 write(*,'(/a)')' Finished!'

contains

 subroutine numlines(unfile,n)
  implicit none
  integer::io
  integer,intent(in)::unfile
  integer,intent(out)::n
  n=0
  do
   read(unfile,*,iostat=io)
   if (io.ne.0) exit
   n=n+1
  enddo
  rewind(unfile)
 end subroutine
 
 subroutine numcol(unfile,n)
  implicit none
  integer,intent(in)::unfile
  character(len=1000000)::a
  integer,intent(out)::n
  integer::curr,first,last,lena,stat,i
  rewind(unfile)
  read(unfile,"(a)")a
  curr=1;lena=len(a);n=0
  do
   first=0
   do i=curr,lena
    if (a(i:i) /= " ") then
     first=i
     exit
    endif
   enddo
   if (first == 0) exit
   curr=first+1
   last=0
   do i=curr,lena
    if (a(i:i) == " ") then
     last=i
     exit
    endif
   enddo
   if (last == 0) last=lena
   n=n+1
   curr=last+1
  enddo
  rewind(unfile)
 end subroutine

 subroutine invsym(a,n)
  !compute the inverse of a real symmetric positive definite matrix
  !using MKL
  implicit none
  integer,intent(in)::n
  integer::info,nmkl,nomp,i,j,mkl_get_max_threads,omp_get_num_threads
  real(8),intent(inout)::a(n,n)

  !$omp parallel
  !$omp master
   nomp=omp_get_num_threads()
  !$omp end master
  !$omp end parallel
   nmkl=mkl_get_max_threads()

  call dpotrf('L',n,a,n,info)
  if (info.ne.0) then
   print*,'ERROR info: ',info
   stop
  endif
  call dpotri('L',n,a,n,info)
  if (info.eq.0) then
!   forall(i=1:n,j=1:n,j>i) a(i,j)=a(j,i) !segmetation fault > 44000
  !$omp parallel
  !$omp do schedule(auto)
   do i=1,n
    do j=i+1,n
     a(i,j)=a(j,i)
    enddo
   enddo
  !$omp enddo
  !$omp end parallel
  else
   print*,'ERROR info: ',info
   stop
  endif
  print*,'Max number of threads used by MKL:',min(nomp,nmkl)

 end subroutine

 real(kind=8) function corr(n,x,y)
  implicit none
  integer,intent(in)::n
  real(kind=8),intent(in)::x(:),y(:)
  integer::i
  real(kind=8)::sumx,sumx2,sumy,sumy2,sumxy2
  real(kind=8)::varx,vary,covxy
  
  sumx=0.d0;sumx2=0.d0;sumy=0.d0;sumy2=0.d0;sumxy2=0.d0
  do i=1,n
   sumx=sumx+x(i)
   sumx2=sumx2+x(i)**2
   sumy=sumy+y(i)
   sumy2=sumy2+y(i)**2
   sumxy2=sumxy2+x(i)*y(i)
  end do
  
  varx=0.d0;vary=0.d0;covxy=0.d0
  varx=(sumx2-(sumx**2)/n)/(n-1.d0)
  vary=(sumy2-(sumy**2)/n)/(n-1.d0)
  covxy=(sumxy2-sumx*sumy/n)/(n-1.d0)
  corr=covxy/sqrt(varx*vary)
 end function


end program
