!ifort -O3 -heap-arrays -lmkl_blas95_lp64 -lmkl_lapack95_lp64 -lmkl_intel_lp64 -lmkl_intel_thread -lmkl_core -qopenmp -parallel -lpthread snpblup+.f90  -o snpblup+

module snpb
 !$ use omp_lib
 implicit none
 private
 
 type::texternal
  private
  integer(kind=4)::next
  character(len=3)::typeinfo,amountinfo                   !typeinfo={pev,pec,r2}  amountinfo={full,chr}
  character(len=30)::files(4)
 end type

 type::tsnpblup
  private
  integer(kind=4)::nan,neffects,nsnp,nsel,next
  integer(kind=4),allocatable::chromo(:)
  real(kind=8)::varg,vare,vars,sum2pq,maf,sde
  real(kind=8),allocatable::genotype(:,:),freq(:),y(:),tbv(:),lhs(:,:),rhs(:),sol(:)
  character(len=30)::title,phenfile,snpfile,varfile,mapfile,parfile,parfile1
  logical,allocatable::selsnp(:)
  logical::lsavemem,linvert,lext
  type(texternal),allocatable::text(:)
  contains
   private
   procedure,public::init
   procedure,public::setuplhsrhssnpblup
   procedure,public::solve
   procedure,public::output
   procedure,public::computeld
 end type



 type(tsnpblup),public::snpblup

contains
 
 !public subroutines
 subroutine init(snp,parfile,parfile1)
  class(tsnpblup),intent(inout)::snp
  character(len=*),intent(in)::parfile
  character(len=*),intent(in),optional::parfile1

  snp%lsavemem=.true.
  snp%linvert=.false.
  snp%lext=.false.
  if(present(parfile1))then
   snp%lext=.true.
   snp%parfile1=adjustl(parfile1(:len_trim(parfile1)))
  endif


  !get parameter files
  snp%parfile=adjustl(parfile(:len_trim(parfile)))
  call readparsnpblup(snp)
  call readparext(snp)

  !read variance file
  call readvariance(snp)

  !read genotype
  call readgenotype(snp)

  !read mapfile
  call readmap(snp)

  !read phenotypes +  tbv
  call readphenotype(snp)

  !compute allele frequencies and remove snp with low maf from genotype matrix
  call allelefrequencies(snp)  

 end subroutine

 subroutine computeld(snp)
  class(tsnpblup),intent(inout)::snp
  integer(kind=4)::i,j,un
  real(kind=8),allocatable::tmp(:)

  !Computation of r2
  open(newunit=un,file='tmpgeno',buffered='yes',access='stream',status='replace')
  write(un)snp%genotype
  close(un)

  !$omp parallel do default(none) shared(snp) private(i)
  do i=1,snp%nsnp
   snp%genotype(:,i)=snp%genotype(:,i)-2.d0*snp%freq(i)
  enddo
  !$omp end parallel do

  allocate(snp%lhs(snp%neffects,snp%neffects))
  call dgemm('t','n',snp%neffects,snp%neffects,snp%nan,1.d0,snp%genotype,snp%nan,snp%genotype,snp%nan,0.d0,snp%lhs,snp%neffects)
 
  deallocate(snp%genotype)
  
  open(newunit=un,file='adjsum_'//adjustl(trim(snp%title))//'.dat',buffered='yes')
  do i=1,snp%nsnp
   write(un,'(i0,3(x,f20.10))')i,snp%lhs(i,i),2*snp%nan*snp%freq(i)*(1.d0-snp%freq(i)),snp%freq(i)
  enddo
  close(un)
 
  allocate(tmp(snp%neffects))
  !$omp parallel default(none) shared(snp,tmp) private(i)
  !$omp do
  do i=1,snp%nsnp
   tmp(i)=sqrt(snp%lhs(i,i))
  enddo
  !$omp end do
  !$omp do
  do i=1,snp%nsnp
   do j=1,snp%nsnp
    if(tmp(i)*tmp(j)>0.d0)then
     snp%lhs(j,i)=(snp%lhs(j,i)/(tmp(i)*tmp(j)))!**2
    else
     snp%lhs(j,i)=0.d0
    endif
   enddo
  enddo
  !$omp end do
  !$omp end parallel 
  open(newunit=un,file='r2_'//adjustl(trim(snp%title))//'.dat',buffered='yes',access='stream')
  write(un)snp%lhs(1:snp%nsnp,1:snp%nsnp)
  close(un)
 
!  if(msg.ne.0)then
!   open(newunit=un,file='r2_ascii_'//adjustl(trim(snp%title))//'.dat',buffered='yes')
!   do i=1,snp%nsnp
!    write(un,'(10000f16.8)')snp%lhs(1:snp%nsnp,i)
!   enddo
!   close(un)
!  endif

  deallocate(snp%lhs)

  !read genotype
  allocate(snp%genotype(snp%nan,snp%neffects))
  open(newunit=un,file='tmpgeno',buffered='yes',access='stream',action='read',status='old')
  read(un)snp%genotype
  close(un,status='delete')

 end subroutine

 subroutine setuplhsrhssnpblup(snp)
  class(tsnpblup),intent(inout)::snp
  integer::i

  !setup variance components
  snp%sde=sqrt(snp%vare)
  write(*,'(/a,f0.4,a)')' Standardisation by sqrt(Vare) [ ',snp%sde,' ]:'
  !sde=1.d0
  !sde=sde/10.d0
  snp%vare=snp%vare/(snp%sde**2)
  snp%varg=snp%varg/(snp%sde**2)
  snp%vars=snp%varg/snp%sum2pq
  write(*,'(a,f0.3)')      ' Heritability                : ',snp%varg/(snp%varg+snp%vare)
  write(*,'(a,f0.3,a,f0.3)')' Genetic / residual variances: ',snp%varg,' / ',snp%vare
  write(*,'(a,e11.4)')      ' SNP variance                : ',snp%vars

  call writevarused(snp)

  snp%vare=1.d0/snp%vare;snp%vars=1.d0/snp%vars

  !setup LHS
  allocate(snp%lhs(snp%neffects,snp%neffects))
  snp%lhs=0.d0
  
  call dgemm('t','n',snp%neffects,snp%neffects,snp%nan,snp%vare,snp%genotype,snp%nan,snp%genotype,snp%nan,0.d0,snp%lhs,snp%neffects)
 
  do i=1,snp%nsnp
   write(555,*)i,snp%lhs(i,i),snp%vars
   snp%lhs(i,i)=snp%lhs(i,i)+snp%vars
  enddo

  !setup RHS
  snp%y=snp%y/snp%sde

  allocate(snp%rhs(snp%neffects))
  open(unit=100,file='scaledpheno'//adjustl(trim(snp%title))//'.dat',buffered='yes')
  write(100,*)
  do i=1,snp%nan
   snp%rhs=0.
   snp%rhs(9)=snp%tbv(i)
   snp%rhs(11)=snp%y(i)
   write(100,'(2(i10,x),9f20.10)')i,i,snp%rhs(3:11)
  enddo
  close(100)

  snp%rhs=0.d0
  call dgemv('t',snp%nan,snp%neffects,snp%vare,snp%genotype,snp%nan,snp%y,1,0.d0,snp%rhs,1)

  deallocate(snp%genotype)

  !external info
  if(snp%lext)then
   write(*,'(/a)')' Start to add external information'
   call addexternal(snp)
  endif

 end subroutine

 subroutine solve(snp)
  class(tsnpblup),intent(inout)::snp

  !$ real(kind=8)::t1

  allocate(snp%sol(snp%neffects))
  snp%sol=0.d0
  write(*,'(/a)')' Inversion of the LHS'
  !$ t1=omp_get_wtime()
  call invsym(snp%lhs,snp%neffects)
  !$ write(*,'(a,f0.3)')'  Elapsed time: ',omp_get_wtime()-t1
  snp%linvert=.true.
  write(*,'(/a)')' Computation of the estimates'
  !$ t1=omp_get_wtime()
  call dgemv('n',snp%neffects,snp%neffects,1.d0,snp%lhs,snp%neffects,snp%rhs,1,0.d0,snp%sol,1)
  !$ write(*,'(a,f0.3)')'  Elapsed time: ',omp_get_wtime()-t1

 end subroutine

 subroutine output(snp)
  class(tsnpblup),intent(inout)::snp
  integer::un,i

  if(allocated(snp%lhs).and.snp%linvert)then      !PEV PEC
   open(newunit=un,file='pev_'//adjustl(trim(snp%title))//'.dat',buffered='yes',access='stream')
   do i=1,snp%neffects
    write(un)snp%lhs(i,i)
   enddo
   close(un)
  
   open(newunit=un,file='pec_'//adjustl(trim(snp%title))//'.dat',buffered='yes',access='stream')
   write(un)snp%lhs
   close(un)
!   if(msg.ne.0)then
!    open(newunit=un,file='pec_ascii_'//adjustl(trim(title))//'.dat',buffered='yes')
!    do i=1,neffects
!     write(un,'(10000f16.8)')lhs(1:neffects,i)
!    enddo
!    close(un)
!   endif
  elseif(allocated(snp%lhs).and..not.snp%linvert)then !LHS
  endif

  if(allocated(snp%sol).and.allocated(snp%lhs))then
   open(newunit=un,file='sol_'//adjustl(trim(snp%title))//'.dat',buffered='yes')
   do i=1,snp%neffects
    write(un,'(i0,4(x,f16.8))')i,snp%sol(i),snp%lhs(i,i),1.d0-(snp%lhs(i,i)*snp%vars),snp%sol(i)*snp%sde
   enddo
   close(un)
   deallocate(snp%lhs,snp%sol,snp%tbv,snp%rhs,snp%freq)
  endif
 
 end subroutine

 !private subroutines
 subroutine readparsnpblup(snp)
  class(tsnpblup),intent(inout)::snp
  integer::un

  open(newunit=un,file=snp%parfile,status='old')
  read(un,*)snp%title
  read(un,*)snp%phenfile
  read(un,*)snp%snpfile
  read(un,*)snp%mapfile
  read(un,*)snp%varfile
  read(un,*)snp%maf
  close(un)
 end subroutine

 subroutine readparext(snp)
  class(tsnpblup),intent(inout)::snp
  integer::un,i,j

  if(snp%lext)then
   open(newunit=un,file=snp%parfile1,status='old')
   read(un,*)snp%next
   allocate(snp%text(snp%next))
   do i=1,snp%next
    read(un,*)snp%text(i)%typeinfo,snp%text(i)%amountinfo,snp%text(i)%files
   enddo
   close(un)
  else
   write(*,'(/a/)')'There is NO external data'
  endif

 end subroutine

 subroutine readvariance(snp)
  class(tsnpblup),intent(inout)::snp
  integer::i,un
  character(len=30)::tmpa

  open(newunit=un,file=snp%varfile,status='old')
  do i=1,4
   read(un,*)
  enddo
  read(un,*)tmpa,snp%varg
  read(un,*)tmpa,snp%vare
  close(un)

  write(*,'(/a,f0.3)')      ' Heritability                : ',snp%varg/(snp%varg+snp%vare)
  write(*,'(a,f0.3,a,f0.3)')' Genetic / residual variances: ',snp%varg,' / ',snp%vare

 end subroutine

 subroutine readmap(snp)
  class(tsnpblup),intent(inout)::snp
  integer::un,i,io,k

  allocate(snp%chromo(snp%nsnp))

  open(newunit=un,file=snp%mapfile,status='old',buffered='yes')
  read(un,*)
  do i=1,snp%nsnp
   read(un,*)k,snp%chromo(i)
  enddo
  close(un)

 end subroutine

 subroutine readgenotype(snp)
  class(tsnpblup),intent(inout)::snp
  integer::un,i,io,k

  open(newunit=un,file=snp%snpfile,status='old',buffered='yes')
  call numlines(un,snp%nan)
  call numcol(un,snp%neffects) 
  snp%nsnp=snp%neffects-1
  allocate(snp%genotype(snp%nan,snp%neffects))
  snp%genotype=-99.d0
  print'(a)',' Start reading the genotype file'
  do i=1,snp%nan
   read(un,*,iostat=io)k,snp%genotype(i,1:snp%nsnp)
   if(io.ne.0)exit
  enddo
  close(un)
  write(*,'(a,i0)')' Number of animals: ',snp%nan
  write(*,'(a,i0)')' Number of SNPs   : ',snp%nsnp
  snp%genotype(:,snp%neffects)=1.d0

 end subroutine

 subroutine readphenotype(snp)
  class(tsnpblup),intent(inout)::snp
  integer::un,i,io,k
  real(kind=8)::rhs(11)

  allocate(snp%y(snp%nan),snp%tbv(snp%nan))
  open(newunit=un,file=snp%phenfile,buffered='yes',status='old')
  read(un,*)
  do i=1,snp%nan
   read(un,*,iostat=io)rhs(:)
   if(io.ne.0)exit
   snp%y(i)=rhs(11)
   snp%tbv(i)=rhs(9)
  enddo
  close(un)

 end subroutine

 subroutine allelefrequencies(snp)
  class(tsnpblup),intent(inout)::snp
  integer::un,i,j,io,k
  real(kind=8)::nalleles,tmp

  print'(/a)',' Computation of allele frequencies'
  allocate(snp%freq(snp%nsnp),snp%selsnp(snp%nsnp))
  snp%freq=0.d0;snp%selsnp=.true.
  nalleles=2.d0*snp%nan
  snp%sum2pq=0.d0
  tmp=0.d0
 
  j=0
  open(newunit=un,file='freq_'//adjustl(trim(snp%title))//'.dat',buffered='yes')
  !$omp parallel default(none) shared(j,nalleles,snp,tmp,un) private(i,k)
  !$omp do reduction(+:tmp,j)
  do i=1,snp%nsnp
   snp%freq(i)=sum(snp%genotype(:,i))/nalleles
   k=(count(snp%genotype(:,i).eq.0))*(count(snp%genotype(:,i).eq.1))*(count(snp%genotype(:,i).eq.2))
   if((snp%freq(i)<=snp%maf.or.snp%freq(i)>=1.d0-snp%maf).or.k.eq.0)then
    snp%selsnp(i)=.false.
    snp%genotype(:,i)=0.d0
    snp%freq(i)=0.d0
    j=j+1
   endif
!   if(lcentered)genotype(:,i)=genotype(:,i)-2.d0*freq(i)
   tmp=tmp+2.d0*snp%freq(i)*(1.d0-snp%freq(i))
   !$omp critical
   write(un,*)i,snp%freq(i),snp%selsnp(i)
   !$omp end critical
  enddo
  !$omp end do
  !$omp end parallel
  close(un)
  snp%nsel=count(snp%selsnp.eq..true.)
  snp%sum2pq=tmp
  write(*,'(a,f0.3)')' Minor allele frequency   : ',snp%maf
  write(*,'(a,i0)')' Number of selected SNPs  : ',snp%nsel
  write(*,'(a,i0)')  ' Number of unselected SNPs: ',j
  write(*,'(a,f0.3)')' Sum of 2pq               : ',snp%sum2pq
  
 end subroutine

 subroutine writevarused(snp)
  class(tsnpblup),intent(inout)::snp
  integer::i,k

  open(newunit=k,file='varused_'//adjustl(trim(snp%title)),buffered='yes',status='replace')
  do i=1,4
   write(k,*)
  enddo
  write(k,*)'varg',snp%varg
  write(k,*)'vare',snp%vare
  write(k,*)'vars',snp%vars
  write(k,*)'nanimals',snp%nan
  write(k,*)'nsnp',snp%nsnp
  write(k,*)'nsel',snp%nsel
  write(k,*)'sum2pq',snp%sum2pq
  close(k)

 end subroutine

 subroutine addexternal(snp)
  class(tsnpblup),intent(inout)::snp
  integer::i,j,k,un,io
  real(kind=8)::tmp1,varsf,rtmp,sum2pq
  real(kind=8),allocatable::pec(:,:),tmp(:),alpha(:),freq(:),totnan(:)
  character(len=30)::atmp

  varsf=0.d0

  allocate(freq(snp%nsnp),totnan(snp%nsnp))
  freq=0.d0;totnan=0.d0
  allocate(pec(snp%nsnp,snp%nsnp),alpha(snp%nsnp))
  do i=1,snp%next
   !alpha
   alpha=0.d0
   open(newunit=un,file=snp%text(i)%files(2),status='old',action='read',buffered='yes')
   do j=1,snp%nsnp
    read(un,*)k,alpha(j)
   enddo
   close(un)

   !variances
   open(newunit=un,file=snp%text(i)%files(3),status='old',action='read',buffered='yes')
   do j=1,6
    read(un,*)
   enddo
   read(un,*)atmp,tmp1
   varsf=varsf+1.d0/tmp1
   read(un,*)atmp,rtmp   !number of animals
   close(un)

   !read allele frequencies
   rtmp=rtmp*2.d0
   open(newunit=un,file=snp%text(i)%files(4),status='old',action='read',buffered='yes')
   do j=1,snp%nsnp
    read(un,*)k,tmp1
    if(tmp1.gt.1e-10)then
     freq(k)=freq(k)+tmp1*rtmp
     totnan(k)=totnan(k)+rtmp   !2*number of total animals
    endif
   enddo
   close(un)

   !lhs
   open(newunit=un,file=snp%text(i)%files(1),access='stream',status='old',action='read',buffered='yes')
   allocate(tmp(snp%nsnp+1))
   tmp=0.d0
   select case(snp%text(i)%typeinfo)
   case('pev')
    read(un)tmp
    pec=0.d0
    do j=1,snp%nsnp
     pec(j,j)=1.d0/tmp(j)
    enddo
   case('pec')
    do j=1,snp%nsnp
     read(un)tmp
     pec(:,j)=tmp(1:snp%nsnp)
    enddo
    if(snp%text(i)%amountinfo.eq.'chr')call zeroing(snp,pec)
    call invsym(pec,snp%nsnp)
   case('r2s')
    !read correlation file
    do j=1,snp%nsnp
     read(un)tmp(1:snp%nsnp)
     pec(:,j)=tmp(1:snp%nsnp)
    enddo
    !if(snp%text(i)%amountinfo.eq.'chr')call zeroing(snp,pec)
    !transform r2 to lhs
    call setuplhsfromr2(snp,i,pec)
   case default
    stop 'this option does not exist'
   end select
   deallocate(tmp)
   close(un)

   !lhs*alpha
   allocate(tmp(snp%nsnp))
   tmp=0.d0
   call dgemv('n',snp%nsnp,snp%nsnp,1.d0,pec,snp%nsnp,alpha,1,1.d0,tmp,1)
   snp%rhs(1:snp%nsnp)=snp%rhs(1:snp%nsnp)+tmp
   snp%lhs(1:snp%nsnp,1:snp%nsnp)=snp%lhs(1:snp%nsnp,1:snp%nsnp)+pec
   deallocate(tmp)
  enddo
  deallocate(pec,alpha)
 
  !new allele frequencies
  freq=freq+snp%freq*snp%nan*2.d0
  totnan=totnan+snp%nan*2.d0

  !compute and write allele frequencies
  sum2pq=0.d0
  open(newunit=un,file='freq_bout_'//adjustl(trim(snp%title)),status='replace',action='write',buffered='yes')
  do j=1,snp%nsnp
   if(freq(j)>0.d0)freq(j)=freq(j)/totnan(j)
   write(un,*)j,freq(j),int(totnan(j)/2.d0)
   sum2pq=sum2pq+totnan(j)*(2.d0*freq(j)*(1-freq(j)))
  enddo
  close(un)
  sum2pq=sum2pq/maxval(totnan)   !it is correct for this implemenation

  !joint variance
  open(newunit=un,file=snp%text(1)%files(3),status='old',action='read',buffered='yes')
  do j=1,4
   read(un,*)
  enddo
  read(un,*)atmp,tmp1 !read varg
  close(un)

  varsf=varsf-sum2pq/tmp1+snp%vars


  !correction for the genetic variance
  open(newunit=k,file='varused_bout_'//adjustl(trim(snp%title)),buffered='yes',status='replace')
  do i=1,4
   write(k,*)
  enddo
  write(k,*)'varg',tmp1
  write(k,*) !'vare',vare
  write(k,*)'vars',tmp1/sum2pq,varsf
  write(k,*)'nanimals',int(maxval(totnan)/2.d0)
  write(k,*)'nsnp',snp%nsnp
  write(k,*)!'nsel',nsel
  write(k,*)'sum2pq',sum2pq
  close(k)

  rtmp=0.d0
  do j=1,snp%nsnp
   snp%lhs(j,j)=snp%lhs(j,j)-varsf
   rtmp=rtmp+snp%lhs(j,j)
  enddo
  varsf=tmp1/sum2pq
  print*,rtmp,rtmp/snp%nsnp
 

  !update
  snp%varg=tmp1
  snp%vars=sum2pq/tmp1
  snp%nan=int(maxval(totnan)/2.d0)
  snp%sum2pq=sum2pq
  snp%freq=freq

 end subroutine
 
 subroutine zeroing(snp,mat)
  class(tsnpblup),intent(inout)::snp
  real(kind=8)::mat(:,:)
  integer::i,j
 
  !$omp parallel do default(none) shared(snp,mat) private(i,j) 
  do i=1,snp%nsnp
   do j=1,snp%nsnp
    if(snp%chromo(i).ne.snp%chromo(j))mat(i,j)=0.d0
   enddo
  enddo
  !$omp end parallel do

 end subroutine

 subroutine setuplhsfromr2(snp,i,r2)
  class(tsnpblup),intent(inout)::snp
  real(kind=8),intent(inout)::r2(:,:)
 
  integer::i,j,k,un
  real(kind=8)::tmp1
  real(kind=8)::rtmp,vare,vs!,sum2pq
  real(kind=8),allocatable::tmp(:),freq(:),totnan(:),diagww(:)
  character(len=30)::atmp
 
 
  allocate(tmp(snp%nsnp),freq(snp%nsnp),totnan(snp%nsnp),diagww(snp%nsnp))
  freq=0.d0;totnan=0.d0!;sum2pq=0.d0
 
!  do i=1,snp%next
   print'(a,i0)','  Computation of LHS from r2 for the external information ',i
   diagww=0.d0
   !individual snp variances
   open(newunit=un,file=snp%text(i)%files(3),status='old',action='read',buffered='yes')
   do j=1,5
    read(un,*)
   enddo
   read(un,*)atmp,vare
   read(un,*)atmp,vs
   read(un,*)atmp,rtmp   !number of animals
   close(un)

   rtmp=rtmp*2.d0

   !freq
   open(newunit=un,file=snp%text(i)%files(4),status='old',action='read',buffered='yes')
   do j=1,snp%nsnp
    read(un,*)k,tmp(k)
    if(tmp(k).gt.1e-10)then
     diagww(k)=sqrt(rtmp*tmp(k)*(1.d0-tmp(k))) !rtmp=2*number of animals
     freq(k)=freq(k)+tmp(k)*rtmp
     totnan(k)=totnan(k)+rtmp   !2*number of total animals
    endif
   enddo
   close(un)

   !setup the matrix W'W (genotype *genotype)
   rtmp=2.d0*rtmp  !rtmp=4*number of animals
   vs=1.d0/vs
   !$omp parallel do default(none) shared(r2,diagww,rtmp,tmp,snp,vare,vs) private(i,j)
   do k=1,snp%nsnp
    do j=1,snp%nsnp
!     r2(k,j)=(sqrt(r2(k,j))*diagww(k)*diagww(j)+2.d0*rtmp*tmp(k)*tmp(j))/vare !rtmp=2*number of animals
     r2(k,j)=(r2(k,j)*diagww(k)*diagww(j)+rtmp*tmp(k)*tmp(j))/vare !rtmp=4*number of animals
    enddo
    r2(k,k)=r2(k,k)+vs     !full lhs of 1 single evaluation
   enddo
   !$omp end parallel do


!   open(newunit=un,file=files(3,i),status='old',action='read',buffered='yes')
!   do j=1,nsnp
!    read(un,*)k,alpha(j)
!   enddo
!   close(un)
!
!   call dgemv('n',nsnp,nsnp,1.d0,r2,nsnp,alpha,1,1.d0,rhs,1)
!
!!   call invsym(r2,nsnp) !no need to invert r2 because it is equal to the lhs
!
!  enddo
  deallocate(freq)
!  deallocate(r2)
! 
!  !compute and write allele frequencies
!  open(newunit=un,file='freq_bout_r2_'//adjustl(trim(t)),status='replace',action='write',buffered='yes')
!  do j=1,nsnp
!   if(freq(j)>0.d0)freq(j)=freq(j)/totnan(j)
!   write(un,*)j,freq(j),int(totnan(j)/2.d0)
!   sum2pq=sum2pq+totnan(j)*(2.d0*freq(j)*(1-freq(j)))
!  enddo
!  close(un)
!  sum2pq=sum2pq/maxval(totnan)   !it is correct for this implemenation  ???????
!
!  !joint variance
!  open(newunit=un,file=files(4,1),status='old',action='read',buffered='yes')
!  do j=1,4
!   read(un,*)
!  enddo
!  read(un,*)atmp,tmp(1) !read varg
!  close(un)
!
!  vars=vars-sum2pq/tmp(1)
!
!  open(newunit=k,file='varused_bout_r2_'//adjustl(trim(title)),buffered='yes',status='replace')
!  do i=1,4
!   write(k,*)
!  enddo
!  write(k,*)'varg',tmp(1)
!  write(k,*)'vare',vare
!  write(k,*)'vars',tmp(1)/sum2pq ,vars
!  write(k,*)'nanimals',int(maxval(totnan)/2.d0)
!  write(k,*)'nsnp',nsnp
!  write(k,*)!'nsel',nsel
!  write(k,*)'sum2pq',sum2pq
!  close(k)
!
!  rtmp=0.d0
!  do j=1,nsnp
!   lhs(j,j)=lhs(j,j)-vars
!   rtmp=rtmp+lhs(j,j)
!  enddo
!  vars=tmp(1)/sum2pq
!  print*,'aaa',rtmp,rtmp/nsnp

 end subroutine


 subroutine numlines(unfile,n)
  implicit none
  integer::io
  integer,intent(in)::unfile
  integer,intent(out)::n
  n=0
  do
   read(unfile,*,iostat=io)
   if (io.ne.0) exit
   n=n+1
  enddo
  rewind(unfile)
 end subroutine
 
 subroutine numcol(unfile,n)
  implicit none
  integer,intent(in)::unfile
  character(len=1000000)::a
  integer,intent(out)::n
  integer::curr,first,last,lena,stat,i
  rewind(unfile)
  read(unfile,"(a)")a
  curr=1;lena=len(a);n=0
  do
   first=0
   do i=curr,lena
    if (a(i:i) /= " ") then
     first=i
     exit
    endif
   enddo
   if (first == 0) exit
   curr=first+1
   last=0
   do i=curr,lena
    if (a(i:i) == " ") then
     last=i
     exit
    endif
   enddo
   if (last == 0) last=lena
   n=n+1
   curr=last+1
  enddo
  rewind(unfile)
 end subroutine

 subroutine invsym(a,n)
  !compute the inverse of a real symmetric positive definite matrix
  !using MKL
  implicit none
  integer,intent(in)::n
  integer::info,nmkl,nomp,i,j,mkl_get_max_threads,omp_get_num_threads
  real(8),intent(inout)::a(n,n)

  !$omp parallel
  !$omp master
   nomp=omp_get_num_threads()
  !$omp end master
  !$omp end parallel
   nmkl=mkl_get_max_threads()

  call dpotrf('L',n,a,n,info)
  if (info.ne.0) then
   print*,'ERROR info: ',info
   stop
  endif
  call dpotri('L',n,a,n,info)
  if (info.eq.0) then
!   forall(i=1:n,j=1:n,j>i) a(i,j)=a(j,i) !segmetation fault > 44000
  !$omp parallel
  !$omp do schedule(auto)
   do i=1,n
    do j=i+1,n
     a(i,j)=a(j,i)
    enddo
   enddo
  !$omp enddo
  !$omp end parallel
  else
   print*,'ERROR info: ',info
   stop
  endif
  print*,'Max number of threads used by MKL:',min(nomp,nmkl)

 end subroutine

end module

program snpblup_p
 !$ use omp_lib
 use snpb
 implicit none
 integer(kind=4)::io,narg
 character(len=100)::parfile,parfile1

 narg=command_argument_count()
 if(narg.lt.1) then
  stop 'Error number argument'
 elseif(narg.eq.1)then
  call get_command_argument(1,value=parfile,status=io)
  call snpblup%init(parfile)
 elseif(narg.eq.2)then
  call get_command_argument(1,value=parfile,status=io)
  call get_command_argument(2,value=parfile1,status=io)
  call snpblup%init(parfile,parfile1)
 endif

 call snpblup%computeld
 call snpblup%setuplhsrhssnpblup
 call snpblup%solve
 call snpblup%output

 write(*,'(a)')'Finished!'

contains

 subroutine numlines(unfile,n)
  implicit none
  integer::io
  integer,intent(in)::unfile
  integer,intent(out)::n
  n=0
  do
   read(unfile,*,iostat=io)
   if (io.ne.0) exit
   n=n+1
  enddo
  rewind(unfile)
 end subroutine
 
 subroutine numcol(unfile,n)
  implicit none
  integer,intent(in)::unfile
  character(len=1000000)::a
  integer,intent(out)::n
  integer::curr,first,last,lena,stat,i
  rewind(unfile)
  read(unfile,"(a)")a
  curr=1;lena=len(a);n=0
  do
   first=0
   do i=curr,lena
    if (a(i:i) /= " ") then
     first=i
     exit
    endif
   enddo
   if (first == 0) exit
   curr=first+1
   last=0
   do i=curr,lena
    if (a(i:i) == " ") then
     last=i
     exit
    endif
   enddo
   if (last == 0) last=lena
   n=n+1
   curr=last+1
  enddo
  rewind(unfile)
 end subroutine



end program
