!ifort -O3 -heap-arrays -lmkl_blas95_lp64 -lmkl_lapack95_lp64 -lmkl_intel_lp64 -lmkl_intel_thread -lmkl_core -qopenmp -parallel -lpthread snpblup.f90  -o snpblup
program snpblup
 !$ use omp_lib
 implicit none
 integer::i,j,k,io,un,nan,neffects,nsnp,nalleles,nsel,nanfixed,nfixed,msg=0
 integer,allocatable::chromo(:)
 character(len=30)::snpfile,mafa,varfile,phenfile,title,mapfile
 real(kind=8)::maf,varg,vare,sum2pq,vars,t1,sde
 real(kind=8),allocatable::genotype(:,:),freq(:),lhs(:,:),rhs(:),y(:),tbv(:)
 logical::loptmem=.true.,lcentered=.false.
 logical,allocatable::selsnp(:)

 !Parameters
 call get_command_argument(1,value=title,status=io)
 call get_command_argument(2,value=snpfile,status=io)
 call get_command_argument(3,value=phenfile,status=io)
 call get_command_argument(4,value=varfile,status=io)
 call get_command_argument(5,value=mafa,status=io)
 read(mafa,'(f16.8)')maf
 call get_command_argument(6,value=mafa,status=io)
 read(mafa,'(i10)')nfixed
 call get_command_argument(7,value=mapfile,status=io)

 if(lcentered)then
  write(*,'(a)')' Genotypes will be centered by 2p'
 else
  write(*,'(a)')' Genotypes will not be centered by 2p'
 endif

 !variances
 open(newunit=un,file=varfile)
 do i=1,4
  read(un,*)
 enddo
 read(un,*)mafa,varg
 read(un,*)mafa,vare
 close(un)

 !Genotypes
 open(newunit=un,file=snpfile,status='old',buffered='yes')
 call numlines(un,nan)
 call numcol(un,neffects) 
 nsnp=neffects-1
 neffects=nsnp+nfixed !number of snp +general mean per pop
 allocate(genotype(nan,neffects),freq(nsnp),selsnp(nsnp))
 genotype=-99.d0
 print'(a)',' Start reading the genotype file'
 do i=1,nan
  read(un,*,iostat=io)k,genotype(i,1:nsnp)
  if(io.ne.0)exit
 enddo
 close(un)
 write(*,'(a,i0)')' Number of animals: ',nan
 write(*,'(a,i0)')' Number of SNPs   : ',nsnp
 nanfixed=nan/nfixed
 write(*,'(a,i0)')' Number of animals per mean: ',nanfixed
 genotype(:,nsnp+1:neffects)=0.d0
 do i=1,nfixed
  genotype((i-1)*nanfixed+1:i*nanfixed,nsnp+i)=1.d0
 enddo


 !Computation of allele frequencies
 print'(/a)',' Computation of allele frequencies'
 freq=0.d0;selsnp=.true.
 nalleles=2.d0*nan
 sum2pq=0.d0

 j=0
 open(unit=100,file='freq_'//adjustl(trim(title))//'.dat',buffered='yes')
 !$omp parallel default(none) shared(j,sum2pq,nsnp,freq,genotype,nalleles,selsnp,maf,lcentered) private(i,k)
 !$omp do reduction(+:sum2pq,j)
 do i=1,nsnp
  freq(i)=sum(genotype(:,i))/nalleles
  k=(count(genotype(:,i).eq.0))*(count(genotype(:,i).eq.1))*(count(genotype(:,i).eq.2))
  if((freq(i)<=maf.or.freq(i)>=1.d0-maf).or.k.eq.0)then
   selsnp(i)=.false.
   genotype(:,i)=0.d0
   freq(i)=0.d0
   j=j+1
  endif
  if(lcentered)genotype(:,i)=genotype(:,i)-2.d0*freq(i)
  sum2pq=sum2pq+2.d0*freq(i)*(1.d0-freq(i))
  !$omp critical
  write(100,*)i,freq(i),selsnp(i)
  !$omp end critical
 enddo
 !$omp end do
 !$omp end parallel
 close(100)
 nsel=count(selsnp.eq..true.)
 write(*,'(a,f0.3)')' Minor allele frequency   : ',maf
 write(*,'(a,i0)')' Number of selected SNPs  : ',nsel
 write(*,'(a,i0)')  ' Number of unselected SNPs: ',j
 write(*,'(a,f0.3)')' Sum of 2pq               : ',sum2pq
 
 deallocate(selsnp)

 vars=varg/sum2pq
 write(*,'(/a,f0.3)')      ' Heritability                : ',varg/(varg+vare)
 write(*,'(a,f0.3,a,f0.3)')' Genetic / residual variances: ',varg,' / ',vare
 write(*,'(a,e11.4)')      ' SNP variance                : ',vars

print*,'aaaaaa'
 write(*,'(/a)')' Standardisation by sqrt(Vare):'
 sde=sqrt(vare)
!sde=1.d0
!sde=sde/10.d0
 vare=vare/(sde**2)
 varg=varg/(sde**2)
 vars=varg/sum2pq
 write(*,'(a,f0.3)')      ' Heritability                : ',varg/(varg+vare)
 write(*,'(a,f0.3,a,f0.3)')' Genetic / residual variances: ',varg,' / ',vare
 write(*,'(a,e11.4)')      ' SNP variance                : ',vars

print*,'aaaaaa'
 open(newunit=k,file='varused_'//adjustl(trim(title)),buffered='yes',status='replace')
 do i=1,4
  write(k,*)
 enddo
 write(k,*)'varg',varg
 write(k,*)'vare',vare
 write(k,*)'vars',vars
 write(k,*)'nanimals',nan
 write(k,*)'nsnp',nsnp
 write(k,*)'nsel',nsel
 write(k,*)'sum2pq',sum2pq
 close(k)

! print*,sum(genotype(:,1)),sum(genotype(:,nsnp))

print*,'aaaaaa'
 !Computation of the lhs
 allocate(lhs(neffects,neffects))
 lhs=0.d0
 
print*,'aaaaaabbbbbb'
 vare=1.d0/vare;vars=1.d0/vars
!vare=1.d0

 open(newunit=un,file='geno.tmp',buffered='yes',access='stream',status='replace')
 write(un)genotype
 close(un)

 do i=1,nan
  genotype(i,:)=genotype(i,:)*sqrt(weight(i))
 enddo

 call dgemm('t','n',neffects,neffects,nan,vare,genotype,nan,genotype,nan,0.d0,lhs,neffects)

 open(newunit=un,file='geno.tmp',buffered='yes',access='stream',status='old')
 read(un)genotype
 close(un,status='delete')

print*,'aaaaaabbbbbb'
 do i=1,nsnp
  lhs(i,i)=lhs(i,i)+vars
!  write(100,*)lhs(i,i)
 enddo

 if(msg>0)then
  open(newunit=un,file='lhs_'//adjustl(trim(title))//'.dat',buffered='yes',access='stream')
  write(un)lhs
  close(un)
  open(newunit=un,file='lhs_ascii_'//adjustl(trim(title))//'.dat',buffered='yes')
  do i=1,neffects
   do j=1,i
    write(un,*)i,j,lhs(j,i)
   enddo
  enddo
  close(un)
  open(newunit=un,file='lhs_a_'//adjustl(trim(title))//'.dat',buffered='yes')
  do i=1,neffects
   write(un,'(10000f20.10)')lhs(:,i)
  enddo
  close(un)
 endif

 !Computation of the rhs
 allocate(rhs(11),y(nan),tbv(nan))
 y=0.d0
 open(newunit=un,file=trim(phenfile),buffered='yes',status='old')
 read(un,*)
 do i=1,nan
  read(un,*,iostat=io)rhs(:)
  if(io.ne.0)exit
  y(i)=rhs(11)
  tbv(i)=rhs(9)
 enddo
 close(un)

 y=y/sde
 
 open(newunit=un,file='scaledpheno'//adjustl(trim(title))//'.dat',buffered='yes')
 write(un,*)
 do i=1,nan
  rhs=0.
  rhs(9)=tbv(i)
  rhs(11)=y(i)
  write(un,'(2(i10,x),9f20.10)')i,i,rhs(3:)
 enddo
 close(un)

 deallocate(rhs)
 allocate(rhs(neffects))
 rhs=0.d0
 call dgemv('t',nan,neffects,vare,genotype,nan,y,1,0.d0,rhs,1)
! print*,rhs(neffects),sum(y)

 !Computation of the solutions
 if(loptmem)then
  !$ t1=omp_get_wtime()
  write(*,'(/a)')' Start writing genotypes to file for saving memory'
  open(newunit=un,file='geno.tmp',buffered='yes',access='stream',status='replace')
  write(un)genotype
  close(un)
  deallocate(genotype)
  !$ write(*,'(a,f0.3)')'  Elapsed time: ',omp_get_wtime()-t1
 endif
 deallocate(y)
 allocate(y(neffects))
 y=0.d0
 write(*,'(a)')' Inversion of the LHS'
 !$ t1=omp_get_wtime()
 call invsym(lhs,neffects)
 !$ write(*,'(a,f0.3)')'  Elapsed time: ',omp_get_wtime()-t1
 write(*,'(a)')' Computation of the estimates'
 !$ t1=omp_get_wtime()
 call dgemv('n',neffects,neffects,1.d0,lhs,neffects,rhs,1,0.d0,y,1)
 !$ write(*,'(a,f0.3)')'  Elapsed time: ',omp_get_wtime()-t1

 open(newunit=un,file='pev_'//adjustl(trim(title))//'.dat',buffered='yes',access='stream')
 do i=1,neffects
  write(un)lhs(i,i)
 enddo
 close(un)

 open(newunit=un,file='pec_'//adjustl(trim(title))//'.dat',buffered='yes',access='stream')
 write(un)lhs
 close(un)

 if(msg.ne.0)then
  open(newunit=un,file='pec_ascii_'//adjustl(trim(title))//'.dat',buffered='yes')
  do i=1,neffects
   write(un,'(10000f16.8)')lhs(1:neffects,i)
  enddo
  close(un)
 endif

 open(newunit=un,file='sol_'//adjustl(trim(title))//'.dat',buffered='yes')
 do i=1,neffects
  write(un,'(i0,3(x,f16.8))')i,y(i),lhs(i,i),1.d0-(lhs(i,i)*vars)
 enddo
 close(un)
 deallocate(lhs)
 
 y=y*sde
 open(newunit=un,file='sola_'//adjustl(trim(title))//'.dat',buffered='yes')
 do i=1,neffects
  write(un,'(i0,x,f16.8)')i,y(i)
 enddo
 close(un)
 
 !Computation of the GEBV
 if(loptmem)then
  !$ t1=omp_get_wtime()
  write(*,'(/a)')' Start reading genotypes from file for saving memory'
  allocate(genotype(nan,neffects))
  open(newunit=un,file='geno.tmp',buffered='yes',access='stream',status='old')
  read(un)genotype
  close(un,status='delete')
  !$ write(*,'(a,f0.3)')'  Elapsed time: ',omp_get_wtime()-t1
 endif
 genotype(:,neffects)=0.d0
 deallocate(rhs)
 allocate(rhs(nan))
 rhs=0.d0
 call dgemv('n',nan,neffects,1.d0,genotype,nan,y,1,0.d0,rhs,1)

 open(newunit=un,file='gebva_'//adjustl(trim(title))//'.dat',buffered='yes')
 do i=1,nan
  write(un,'(i0,x,f16.8)')i,rhs(i)
 enddo
 close(un)

 !Computation of correlation (accuracy)
 write(*,'(/a,f0.3)')' Correlation between GEBV and TBV: ',corr(nan,rhs,tbv)
 deallocate(rhs,tbv)

 !Computation of r2
 if(.not.lcentered)then
  !$omp parallel do default(none) shared(nsnp,genotype,freq) private(i)
  do i=1,nsnp
   genotype(:,i)=genotype(:,i)-2.d0*freq(i)
  enddo
  !$omp end parallel do
 endif
 allocate(lhs(neffects,neffects))
 call dgemm('t','n',neffects,neffects,nan,1.d0,genotype,nan,genotype,nan,0.d0,lhs,neffects)
 
 deallocate(genotype)
 open(unit=100,file='adjsum_'//adjustl(trim(title))//'.dat',buffered='yes')
 do i=1,nsnp
  write(100,'(i0,3(x,f20.10))')i,lhs(i,i),2*nan*freq(i)*(1.d0-freq(i)),freq(i)
 enddo
 close(100)

 !$omp parallel default(none) shared(nsnp,freq,lhs) private(i)
 !$omp do
 do i=1,nsnp
  freq(i)=sqrt(lhs(i,i))
 enddo
 !$omp end do
 !$omp do
 do i=1,nsnp
  do j=1,nsnp
   if(freq(i)*freq(j)>0.d0)then
    lhs(j,i)=(lhs(j,i)/(freq(i)*freq(j)))!**2
   else
    lhs(j,i)=0.d0
   endif
  enddo
 enddo
 !$omp end do
 !$omp end parallel 
 open(newunit=un,file='r2_'//adjustl(trim(title))//'.dat',buffered='yes',access='stream')
 write(un)lhs(1:nsnp,1:nsnp)
 close(un)

 if(msg.ne.0)then
  open(newunit=un,file='r2_ascii_'//adjustl(trim(title))//'.dat',buffered='yes')
  do i=1,nsnp
   write(un,'(10000f16.8)')lhs(1:nsnp,i)
  enddo
  close(un)
 endif

! !read map file
! allocate(chromo(nsnp))
! chromo=0
! open(newunit=un,file=mapfile,buffered='yes')
! read(un,*)
! do i=1,nsnp
!  read(un,*)k,chromo(i)
! enddo
! close(un)
!
!
! !$omp parallel do default(none) shared(chromo,nsnp,lhs) private(i,j)
! do i=1,nsnp
!  do j=1,nsnp
!   if(chromo(i).ne.chromo(j))then
!    lhs(i,j)=0.d0
!   endif
!  enddo
! enddo
! !$omp end parallel do
!
! open(newunit=un,file='r2c_'//adjustl(trim(title))//'.dat',buffered='yes',access='stream')
! write(un)lhs(1:nsnp,1:nsnp)
! close(un)

! deallocate(y,lhs,freq,chromo)

 deallocate(y,lhs,freq)

 write(*,'(a)')'Finished!'

contains

 subroutine numlines(unfile,n)
  implicit none
  integer::io
  integer,intent(in)::unfile
  integer,intent(out)::n
  n=0
  do
   read(unfile,*,iostat=io)
   if (io.ne.0) exit
   n=n+1
  enddo
  rewind(unfile)
 end subroutine
 
 subroutine numcol(unfile,n)
  implicit none
  integer,intent(in)::unfile
  character(len=1000000)::a
  integer,intent(out)::n
  integer::curr,first,last,lena,stat,i
  rewind(unfile)
  read(unfile,"(a)")a
  curr=1;lena=len(a);n=0
  do
   first=0
   do i=curr,lena
    if (a(i:i) /= " ") then
     first=i
     exit
    endif
   enddo
   if (first == 0) exit
   curr=first+1
   last=0
   do i=curr,lena
    if (a(i:i) == " ") then
     last=i
     exit
    endif
   enddo
   if (last == 0) last=lena
   n=n+1
   curr=last+1
  enddo
  rewind(unfile)
 end subroutine

 subroutine invsym(a,n)
  !compute the inverse of a real symmetric positive definite matrix
  !using MKL
  implicit none
  integer,intent(in)::n
  integer::info,nmkl,nomp,i,j,mkl_get_max_threads,omp_get_num_threads
  real(8),intent(inout)::a(n,n)

  !$omp parallel
  !$omp master
   nomp=omp_get_num_threads()
  !$omp end master
  !$omp end parallel
   nmkl=mkl_get_max_threads()

  call dpotrf('L',n,a,n,info)
  if (info.ne.0) then
   print*,'ERROR info: ',info
   stop
  endif
  call dpotri('L',n,a,n,info)
  if (info.eq.0) then
!   forall(i=1:n,j=1:n,j>i) a(i,j)=a(j,i) !segmetation fault > 44000
  !$omp parallel
  !$omp do schedule(auto)
   do i=1,n
    do j=i+1,n
     a(i,j)=a(j,i)
    enddo
   enddo
  !$omp enddo
  !$omp end parallel
  else
   print*,'ERROR info: ',info
   stop
  endif
  print*,'Max number of threads used by MKL:',min(nomp,nmkl)

 end subroutine

 real(kind=8) function corr(n,x,y)
  implicit none
  integer,intent(in)::n
  real(kind=8),intent(in)::x(:),y(:)
  integer::i
  real(kind=8)::sumx,sumx2,sumy,sumy2,sumxy2
  real(kind=8)::varx,vary,covxy
  
  sumx=0.d0;sumx2=0.d0;sumy=0.d0;sumy2=0.d0;sumxy2=0.d0
  do i=1,n
   sumx=sumx+x(i)
   sumx2=sumx2+x(i)**2
   sumy=sumy+y(i)
   sumy2=sumy2+y(i)**2
   sumxy2=sumxy2+x(i)*y(i)
  end do
  
  varx=0.d0;vary=0.d0;covxy=0.d0
  varx=(sumx2-(sumx**2)/n)/(n-1.d0)
  vary=(sumy2-(sumy**2)/n)/(n-1.d0)
  covxy=(sumxy2-sumx*sumy/n)/(n-1.d0)
  corr=covxy/sqrt(varx*vary)
 end function


end program
