!ifort -O3 -heap-arrays -lmkl_blas95_lp64 -lmkl_lapack95_lp64 -lmkl_intel_lp64 -lmkl_intel_thread -lmkl_core -qopenmp -parallel -lpthread computeERC.f90  -o computeERC

module snpb
 !$ use omp_lib
 implicit none
 private
 
 type::texternal
  private
  integer(kind=4)::next,amountinfo !amountinfo=Number of SNPs for computation of ERC
  character(len=3)::typeinfo             !typeinfo={pev,pec,r2}  
  character(len=100)::files(4)
 end type

 type::tsnpblup
  private
  integer(kind=4)::nan,neffects,nsnp,nsel,next
  integer(kind=4),allocatable::chromo(:)
  real(kind=8)::varg,vare,vars,sum2pq,maf,sde
  real(kind=8),allocatable::genotype(:,:),freq(:),y(:),tbv(:),lhs(:,:),rhs(:),sol(:),weight(:)
  character(len=100)::title,phenfile,snpfile,varfile,mapfile,parfile,parfile1
  logical,allocatable::selsnp(:)
  logical::lsavemem,linvert,lext
  type(texternal),allocatable::text(:)
  contains
   private
   procedure,public::init
   procedure,public::setuplhsrhssnpblup
 end type



 type(tsnpblup),public::snpblup

contains
 
 !public subroutines
 subroutine init(snp,parfile1)
  class(tsnpblup),intent(inout)::snp
  character(len=*),intent(in)::parfile1

  snp%lsavemem=.true.
  snp%linvert=.false.

  snp%lext=.true.
  snp%parfile1=adjustl(parfile1(:len_trim(parfile1)))

  !get parameter files
  call readparext(snp)

!  !read mapfile
!  call readmap(snp)

 end subroutine

 subroutine setuplhsrhssnpblup(snp)
  class(tsnpblup),intent(inout)::snp

  integer::i,un
  real(kind=8),allocatable::rdummy(:)

  do i=1,snp%next 
   write(*,'(/a,i0)')' Start to add external information ',i
   call computeerc(snp,i)
  enddo

 end subroutine

 !private subroutines
 subroutine readparext(snp)
  class(tsnpblup),intent(inout)::snp
  integer::un,i,j

  open(newunit=un,file=snp%parfile1,status='old')
  read(un,*)snp%maf
  read(un,*)snp%next
  allocate(snp%text(snp%next))
  do i=1,snp%next
   read(un,*)snp%text(i)%typeinfo,snp%text(i)%amountinfo,snp%text(i)%files
  enddo
  close(un)

 end subroutine

 subroutine readmap(snp)
  class(tsnpblup),intent(inout)::snp
  integer::un,i,io,k

  allocate(snp%chromo(snp%nsnp))

  open(newunit=un,file=snp%mapfile,status='old',buffered='yes')
  read(un,*)
  do i=1,snp%nsnp
   read(un,*)k,snp%chromo(i)
  enddo
  close(un)

 end subroutine

 subroutine computeerc(snp,info)
  class(tsnpblup),intent(inout)::snp
  integer,intent(in)::info

  integer::j,k,un
  real(kind=8)::vare,varsf,tmp1,rtmp
  real(kind=8),allocatable::freq(:),tmp(:),pev(:),r2(:,:)
  character(len=30)::atmp

  vare=0
  varsf=0.d0

  !variances
  open(newunit=un,file=snp%text(info)%files(2),status='old',action='read',buffered='yes')
  do j=1,5
   read(un,*)
  enddo
  read(un,*)atmp,vare     !residual variance
  vare=1.0/vare
  read(un,*)atmp,tmp1     !genetic variance
  varsf=varsf+1.d0/tmp1
  read(un,*)atmp,rtmp   !number of animals
  read(un,*)atmp,snp%nsnp   !number of snps
  close(un)
 
  !read r2 (if present)
  print*,'r2'
  if(snp%text(info)%files(4).ne.'dum')then
   allocate(r2(snp%nsnp,snp%nsnp))
   r2=0.d0
   open(newunit=un,file=snp%text(info)%files(4),status='old',action='read',access='stream',buffered='yes')
   read(un)r2
   close(un)
  endif

  !read allele frequencies
  print*,'freq'
  allocate(freq(snp%nsnp))
  freq=0.d0
  open(newunit=un,file=snp%text(info)%files(3),status='old',action='read',buffered='yes')
  do j=1,snp%nsnp
   read(un,*)k,tmp1
   if(tmp1.gt.1e-10)then    !maf?????????????????????????
    freq(k)=tmp1
   endif
  enddo
  close(un)

  !PEV/PEC
  print*,'pev'
  open(newunit=un,file=snp%text(info)%files(1),access='stream',status='old',action='read',buffered='yes')
  allocate(tmp(snp%nsnp+1))
  tmp=0.d0
  select case(snp%text(info)%typeinfo)
  case('pev')
   allocate(pev(snp%nsnp))
   pev=0.d0
   read(un)tmp
   do j=1,snp%nsnp
    pev(j)=tmp(j)
    if(freq(j).ge.snp%maf)then
     tmp(j)=((1.d0/pev(j))-varsf)/(vare*(4*freq(j)*freq(j)+2*freq(j)*(1.d0-freq(j))))
!print*,pev(j),(1.d0/pev(j)),varsf
    else
     tmp(j)=0.d0
    endif
   enddo
  case('r2')
   allocate(pev(snp%nsnp))
   pev=0.d0
   read(un)tmp
   do j=1,snp%nsnp
    pev(j)=tmp(j)
    if(freq(j).ge.snp%maf)then
    else
     tmp(j)=0.d0
    endif
   enddo
   deallocate(tmp);allocate(tmp(snp%nsnp))
   call computeerc_algo(snp,info,pev,freq,r2,vare,varsf,tmp)
!   do j=1,snp%text(info)%amountinfo
!    write(500,*)j,pev(j),freq(j)
!    write(501,'(100(x,f0.8))')r2(1:100,j)
!   enddo

!  case('pec')
!   do j=1,snp%nsnp
!    read(un)tmp
!    pec(:,j)=tmp(1:snp%nsnp)
!   enddo
!   if(snp%text(i)%amountinfo.eq.'chr')call zeroing(snp,pec)
!   call invsym(pec,snp%nsnp)
  case default
   stop 'this option does not exist'
  end select
  close(un)

  !write results
  open(newunit=un,file='erc'//snp%text(info)%files(1),status='replace',action='write',buffered='yes')
  do j=1,snp%nsnp
   write(un,*)j,tmp(j)
  enddo
  close(un)
  deallocate(tmp)

 end subroutine
 
 subroutine zeroing(snp,mat)
  class(tsnpblup),intent(inout)::snp
  real(kind=8)::mat(:,:)
  integer::i,j
 
  !$omp parallel do default(none) shared(snp,mat) private(i,j) 
  do i=1,snp%nsnp
   do j=1,snp%nsnp
    if(snp%chromo(i).ne.snp%chromo(j))mat(i,j)=0.d0
   enddo
  enddo
  !$omp end parallel do

 end subroutine

 subroutine computeerc_algo(snp,info,pev,freq,r2,vare,varsf,weight)
  class(tsnpblup),intent(inout)::snp
  integer::info
  real(kind=8)::pev(:),freq(:),r2(:,:),vare,varsf,weight(:)

  integer::i,j,k,m,iter=2000,nstart,nend,nlength
  real(kind=8)::conv
  real(kind=8),allocatable::V(:),WW(:,:),re0(:),WWsub(:,:),tmp1(:),tmp(:),s(:),q(:),lhs(:,:)
 
  print*,'Start compute ERC ',info
  allocate(V(snp%nsnp))!,WW(snp%nsnp,snp%nsnp))
  V=0.d0

  do i=1,snp%nsnp
   V(i)=sqrt(2.d0*freq(i)*(1.d0-freq(i)))
  enddo

!  WW=r2
!  do i=1,snp%nsnp
!   do j=1,snp%nsnp
!    WW(i,j)=WW(i,j)*V(i)*V(j)+4*freq(i)*freq(j)
!   enddo
!   if(WW(i,i).eq.0.d0)then
!    WW(i,i)=1.d0
!   endif
!  enddo

  allocate(re0(snp%nsnp))
!  do i=1,snp%nsnp
!   re0(i)=((1.d0/pev(i))-varsf)/(vare*WW(i,i))
!  enddo
 

  do j=1,snp%nsnp,snp%text(info)%amountinfo
   nstart=j
   nend=min(snp%nsnp,j+snp%text(info)%amountinfo-1)
   nlength=nend-nstart+1
   print*,'aaa',j,nstart,nend,nlength,snp%nsnp,j+snp%text(info)%amountinfo,-1

   allocate(lhs(nlength,nlength),WWsub(nlength,nlength),tmp1(nlength),q(nlength),s(nlength),tmp(nlength))

   !WWsub=WW(nstart:nend,nstart:nend)
   WWsub=r2(nstart:nend,nstart:nend)
   do k=1,nlength
    do m=1,nlength
     WWsub(k,m)=WWsub(k,m)*V(nstart+k-1)*V(nstart+m-1)+4*freq(nstart+k-1)*freq(nstart+m-1)
    enddo
    if(WWsub(k,k).eq.0.d0)then
     WWsub(k,k)=1.d0
    endif
   enddo

   do k=1,nlength
    re0(k+nstart-1)=((1.d0/pev(k+nstart-1))-varsf)/(vare*WWsub(k,k))
    if(re0(k+nstart-1)<0)re0(k+nstart-1)=0.d0
   enddo

   !Q=diag(sqrt(re0));
   q=re0(nstart:nend)

   do i=1,iter
    ! lhs=Q'*WW*Q*varei+Bi*varai;
    do k=1,nlength
     do m=1,nlength
      lhs(k,m)=WWsub(k,m)*sqrt(q(k))*sqrt(q(m))*vare
     enddo
     lhs(k,k)=lhs(k,k)+varsf
    enddo
! do k=1,nend
!  write(503,'(100(x,f0.8))')lhs(1:nend,k)
! enddo
    !lhs=inv(lhs)
    call invsym(lhs,nlength)
    ! pev=diag(lhs^-1));
    do k=1,nlength
     tmp1(k)=lhs(k,k)
    enddo
    ! re=((1./pev)-diag(Bi*varai))*vare./ww_d;
    do k=1,nlength
     tmp(k)=((1.d0/tmp1(k))-varsf)/(vare*WWsub(k,k))
    enddo
    !% res=[res re];
    ! s=re0-re;
     s=re0(nstart:nend)-tmp
    ! conv=sum(abs(s));
    conv=sum(abs(s));
    ! [i conv]
    print*,i,conv,conv/sum(re0(nstart:nend)),conv/real(nend-nstart+1)
    ! re=diag(Q'*Q)+s;
    q=q+s
    if(conv/real(nend-nstart+1).lt.1.e-1)exit
    ! d=find(re<0);
    ! re(d)=0;
    do k=1,nlength
     if(q(k)<0.d0)q(k)=0.d0
    enddo
    ! Q=diag(sqrt(re));
    ! rr=[i;re]';
    ! res=[res; rr];
   enddo
   weight(nstart:nend)=q
   deallocate(lhs,WWsub,tmp1,q,s,tmp)
  enddo
     

 end subroutine

 subroutine numlines(unfile,n)
  implicit none
  integer::io
  integer,intent(in)::unfile
  integer,intent(out)::n
  n=0
  do
   read(unfile,*,iostat=io)
   if (io.ne.0) exit
   n=n+1
  enddo
  rewind(unfile)
 end subroutine
 
 subroutine numcol(unfile,n)
  implicit none
  integer,intent(in)::unfile
  character(len=1000000)::a
  integer,intent(out)::n
  integer::curr,first,last,lena,stat,i
  rewind(unfile)
  read(unfile,"(a)")a
  curr=1;lena=len(a);n=0
  do
   first=0
   do i=curr,lena
    if (a(i:i) /= " ") then
     first=i
     exit
    endif
   enddo
   if (first == 0) exit
   curr=first+1
   last=0
   do i=curr,lena
    if (a(i:i) == " ") then
     last=i
     exit
    endif
   enddo
   if (last == 0) last=lena
   n=n+1
   curr=last+1
  enddo
  rewind(unfile)
 end subroutine

 subroutine invsym(a,n)
  !compute the inverse of a real symmetric positive definite matrix
  !using MKL
  implicit none
  integer,intent(in)::n
  integer::info,nmkl,nomp,i,j,mkl_get_max_threads,omp_get_num_threads
  real(8),intent(inout)::a(n,n)

  !$omp parallel
  !$omp master
   nomp=omp_get_num_threads()
  !$omp end master
  !$omp end parallel
   nmkl=mkl_get_max_threads()

  call dpotrf('L',n,a,n,info)
  if (info.ne.0) then
   print*,'ERROR info: ',info
   stop
  endif
  call dpotri('L',n,a,n,info)
  if (info.eq.0) then
!   forall(i=1:n,j=1:n,j>i) a(i,j)=a(j,i) !segmetation fault > 44000
  !$omp parallel
  !$omp do schedule(auto)
   do i=1,n
    do j=i+1,n
     a(i,j)=a(j,i)
    enddo
   enddo
  !$omp enddo
  !$omp end parallel
  else
   print*,'ERROR info: ',info
   stop
  endif
!  print*,'Max number of threads used by MKL:',min(nomp,nmkl)

 end subroutine

end module

program snpblup_p
 !$ use omp_lib
 use snpb
 implicit none
 integer(kind=4)::io,narg
 character(len=100)::parfile,parfile1

 narg=command_argument_count()
 if(narg.lt.1) then
  stop 'Error number argument'
 elseif(narg.eq.1)then
  call get_command_argument(1,value=parfile,status=io)
  call snpblup%init(parfile)
 endif

 call snpblup%setuplhsrhssnpblup

 write(*,'(a)')'Finished!'

contains

 subroutine numlines(unfile,n)
  implicit none
  integer::io
  integer,intent(in)::unfile
  integer,intent(out)::n
  n=0
  do
   read(unfile,*,iostat=io)
   if (io.ne.0) exit
   n=n+1
  enddo
  rewind(unfile)
 end subroutine
 
 subroutine numcol(unfile,n)
  implicit none
  integer,intent(in)::unfile
  character(len=1000000)::a
  integer,intent(out)::n
  integer::curr,first,last,lena,stat,i
  rewind(unfile)
  read(unfile,"(a)")a
  curr=1;lena=len(a);n=0
  do
   first=0
   do i=curr,lena
    if (a(i:i) /= " ") then
     first=i
     exit
    endif
   enddo
   if (first == 0) exit
   curr=first+1
   last=0
   do i=curr,lena
    if (a(i:i) == " ") then
     last=i
     exit
    endif
   enddo
   if (last == 0) last=lena
   n=n+1
   curr=last+1
  enddo
  rewind(unfile)
 end subroutine



end program
